#!/usr/bin/env python3

# -*- coding: UTF-8 -*-

"""
Description: The goal of this script is to see if the nucleotide \
frequency is not random between communities
"""


import pandas as pd
import logging
import sqlite3
from .config import ConfigGraph, get_communities
from typing import List
from doctest import testmod
from functools import reduce
from pathlib import Path
from rpy2.robjects import r, pandas2ri
from statsmodels.stats.multitest import multipletests
import numpy as np
from .community_finder import get_projects_name
from ..logging_conf import logging_def
from itertools import product
import multiprocessing as mp
from .sf_and_communities import get_key


def get_nt_frequency(cnx: sqlite3.Connection, list_ft: List[str],
                     feature: str, region: str = "") -> pd.DataFrame:
    """
    Get the frequency of every nucleotides for features in list_ft.

    :param cnx: Connection to chia-pet database
    :param list_ft: The list of exons for which we want to get
    :param feature: the kind of feature analysed
    :param region: The region of gene analysed if feature is gene
    :return: the frequency of nucleotides for the list of exons.

    >>> d = get_nt_frequency(sqlite3.connect(ConfigGraph.db_file),
    ... ["1_1", "1_2"], "exon")
    >>> d[["id_exon", 'A', 'C', 'G', 'T']]
    ft id_exon         A         C         G         T
    0      1_1  16.63480  34.60803  32.12237  16.63480
    1      1_2  16.06426  26.10442  39.75904  18.07229
    >>> d = get_nt_frequency(sqlite3.connect(ConfigGraph.db_file),
    ... ['1', '2'], "gene")
    >>> d[["id_gene", 'A', 'C', 'G', 'T']]
    ft id_gene         A         C         G         T
    0        1  29.49376  18.34271  18.43874  33.72479
    1        2  31.90401  16.40251  18.79033  32.90315
    """
    query_region = ""
    if feature == "gene":
        list_ft = [int(ft) for ft in list_ft]
        if region == "":
            region = "gene"
        query_region = f"AND region = '{region}'"
    query = f"""
             SELECT ft, id_{feature}, frequency
             FROM cin_{feature}_frequency
             WHERE id_{feature} IN {tuple(list_ft)}
             AND ft_type="nt" 
             {query_region}
             """
    df = pd.read_sql_query(query, cnx)
    df = df.pivot_table(index=f"id_{feature}", columns="ft", values="frequency")\
        .reset_index()
    df[f"id_{feature}"] = df[f"id_{feature}"].astype(str)
    return df


def get_community_table(communities: List[List[str]],
                        size_threshold: int, feature: str) -> pd.DataFrame:
    """
    return the table indicating the name of the exons and the \
    the name of the community.

    :param communities: List of community of exons
    :param size_threshold: The required size a community must \
    have to be considered
    :param feature: The kind of feature analysed
    :return: table of community
    >>> c = [['1_1', '2_5'], ['7_9', '4_19', '3_3']]
    >>> get_community_table(c, 3, 'exon')
      community id_exon  community_size
    0        C1     7_9               3
    1        C1    4_19               3
    2        C1     3_3               3
    >>> c = [['1', '2'], ['7', '49', '3']]
    >>> get_community_table(c, 3, 'gene')
      community id_gene  community_size
    0        C1       7               3
    1        C1      49               3
    2        C1       3               3
    """
    dic = {"community": [], f"id_{feature}": [], "community_size": []}
    for k, comm in enumerate(communities):
        if len(comm) >= size_threshold:
            name = f'C{k}'
            clen = len(comm)
            for exon in comm:
                dic["community"].append(name)
                dic[f'id_{feature}'].append(exon)
                dic["community_size"].append(clen)
    return pd.DataFrame(dic)


def lmm_maker(df: pd.DataFrame, outfile: Path, nt: str) -> float:
    """
    Make the lmm analysis to see if the exon regulated by a splicing factor \
    are equally distributed among the communities.

    :param df: The dataframe
    :param outfile: A name of a file
    :param nt: the nucleotide of interest
    :return: the pvalue of lmm

    """
    pandas2ri.activate()
    lmm = r(
        """
        require("lme4")
        require("DHARMa")

        function(data, folder, partial_name) {
            null_mod <- lm(%s ~ log(community_size), data=data)
            mod <- lmer(%s ~ log(community_size) + (1 | community), data=data)
            simulationOutput <- simulateResiduals(fittedModel = mod, n = 250)
            png(paste0(folder, "/dignostics_", partial_name, ".png"))
            plot(simulationOutput)
            dev.off()
            return(anova(mod, null_mod, test="Chisq"))
        }

        """ % (nt, nt))
    folder = outfile.parent / "diagnostics"
    folder.mkdir(parents=True, exist_ok=True)
    partial_name = outfile.name.replace('.txt', '')
    res = lmm(df, str(folder), partial_name)
    return res["Pr(>Chisq)"][1]


def lmm_maker_summary(df: pd.DataFrame, outfile: Path, nt: str
                      ) -> pd.DataFrame:
    """
    Make the lmm analysis to see if the exon regulated by a splicing factor \
    are equally distributed among the communities.

    :param df: The dataframe
    :param outfile: A name of a file
    :param nt: the nucleotide of interest
    :return: the pvalue of lmm

    """
    pandas2ri.activate()
    lmm = r(
        """
        require("DHARMa")

        function(data, folder, partial_name) {
            mod <- lm(%s ~ log(community_size) +  community, data=data)
            simulationOutput <- simulateResiduals(fittedModel = mod, n = 250)
            png(paste0(folder, "/dignostics_summary", partial_name, ".png"))
            plot(simulationOutput)
            dev.off()
            return(as.data.frame(summary(mod)$coefficients))
        }

        """ % nt)
    folder = outfile.parent / "diagnostics"
    folder.mkdir(parents=True, exist_ok=True)
    partial_name = outfile.name.replace('.txt', '')
    res_df = lmm(df, str(folder), partial_name).reset_index()
    res_df.rename({'index': 'community'}, inplace=True, axis=1)
    res_df['community'] = res_df['community'].str.replace('community', '')
    res_df.loc[res_df['community'] == "(Intercept)", "community"] = "C-CTRL"
    mean_df = df[[nt, "community", "community_size"]].\
        groupby(["community", "community_size"]).mean().reset_index()
    return res_df.merge(mean_df, how="left", on="community")


def get_ft_id(cnx: sqlite3.Connection, feature: str = "exon") -> List[str]:
    """
    Return the id of every gene/exons in chia-pet database.

    :param cnx: A connection to chiapet database
    :param feature: The feature of interest
    :return: The list of feature id
    """
    query = f"SELECT DISTINCT id FROM cin_{feature}"
    c = cnx.cursor()
    c.execute(query)
    res = c.fetchall()
    return [str(cid[0]) for cid in res]


def create_ctrl_community(df: pd.DataFrame, outfile: Path,
                          feature: str = 'exon', region: str = ""
                          ) -> pd.DataFrame:
    """
    :param df: A dataframe containing the frequency of each feature in a \
    community.
    :param outfile: The output table containing frequencies
    :param feature: The kind of feature to analyse
    :return: A dataframe containing the frequency of every nucleotides \
    of every exon in a large community
    """
    if outfile.is_file():
        return pd.read_csv(outfile, sep="\t")
    size_threshold = 10 if feature == "gene" else 50
    cnx = sqlite3.connect(ConfigGraph.db_file)
    ft_id = get_ft_id(cnx, feature)
    list_ft = [cid for cid in ft_id
               if cid not in df[f"id_{feature}"].to_list()]
    df_nt = get_nt_frequency(cnx, list_ft, feature, region)
    df_com = get_community_table([list_ft], size_threshold, feature)
    df_nt = df_nt.merge(df_com, how="left", on=f"id_{feature}")
    df_nt['community'] = ['C-CTRL'] * df_nt.shape[0]
    df = pd.concat([df, df_nt], axis=0, ignore_index=True)
    df.to_csv(outfile, sep="\t", index=False)
    return df


def get_stat_nt_communities(df: pd.DataFrame, project: str, weight: int,
                            global_weight: int, same_gene: bool,
                            nt: str, feature: str = "exon",
                            region: str = "",
                            ) -> pd.Series:
    """
    Get data (proportion of `reg` regulated exons by a splicing factor
    `sf_name` within a community) for every community.

    :param df: A dataframe containing the frequency of each nucleotide \
    in each exons belonging to a community.
    :param project: The name of the project of interest
    :param weight: The minimum weight of interaction to consider
    :param global_weight: The global weight to consider. if \
    the global weight is equal to 0 then then density figure are calculated \
    by project, else all projet are merge together and the interaction \
    seen in `global_weight` project are taken into account
    :param same_gene: Say if we consider as co-localised, exons within the \
    same gene (True) or not (False) (default False)
    :param nt: The nucleotide of interest
    :param feature: The king of feature analysed
    :param region: the region of interest to extract from gene
    """
    logging.debug(f"lmm for {project}, w:{weight}, "
                  f"g:{global_weight} nt: {nt}")
    res = {"project": [], "nt": [], 'pval': []}

    outfile = ConfigGraph.get_community_file(project, weight,
                                             global_weight,
                                             same_gene, feature,
                                             f"{nt}_stat.txt",
                                             True)
    nfolder = outfile.parent / "nt_analysis"
    nfolder.mkdir(exist_ok=True, parents=True)
    noutfile = nfolder / outfile.name
    pval = lmm_maker(df, noutfile, nt)
    res['project'] = project
    res['nt'] = nt
    res['pval'] = pval
    nt_ctrl_table = noutfile.parent / noutfile.name.replace("_stat.txt",
                                                            "_ctrl.txt")
    ndf = create_ctrl_community(df, nt_ctrl_table, feature, region)
    sum_df = lmm_maker_summary(ndf, outfile, nt)
    outfile_ctrl = ConfigGraph.get_community_file(project, weight,
                                                  global_weight,
                                                  same_gene, feature,
                                                  f"{nt}_VS_CTRL_stat.txt",
                                                  True)
    noutfile_ctrl = nfolder / outfile_ctrl.name
    sum_df.to_csv(noutfile_ctrl, sep="\t", index=False)
    return pd.Series(res)


def create_dataframe(project: str, weight: int, global_weight: int,
                     same_gene: bool, feature: str = 'exon') -> pd.DataFrame:
    """
    :param project: The name of the project of interest
    :param weight: The minimum weight of interaction to consider
    :param global_weight: The global weight to consider. if \
    the global weight is equal to 0 then then density figure are calculated \
    by project, else all projet are merge together and the interaction \
    seen in `global_weight` project are taken into account
    :param same_gene: Say if we consider as co-localised, exons within the \
    same gene (True) or not (False)
    :param feature: The kind of feature to analyse
    :return: A dataframe containing the frequency of every nucleotides \
    of every exon in a large community
    """
    size_threshold = 10 if feature == "gene" else 50
    result = ConfigGraph.get_community_file(project, weight, global_weight,
                                            same_gene, feature,
                                            "_communities.txt")
    communities = get_communities(result, 0)
    ncommunities = [c for c in communities if len(c) >= size_threshold]
    cnx = sqlite3.connect(ConfigGraph.db_file)
    list_exon = reduce(lambda x, y: x + y, ncommunities)
    df = get_nt_frequency(cnx, list_exon, feature)
    df_com = get_community_table(communities, size_threshold, feature)
    df = df.merge(df_com, how="left", on=f"id_{feature}")
    return df


def multiple_nt_lmm_launcher(ps: int, weights: List[int],
                             global_weights: List[int],
                             same_gene: bool,
                             feature: str = 'exon', region: str = '',
                             logging_level: str = "DISABLE"):
    """
    Launch the statistical analysis for every

    :param ps: The number of processes we want to use.
    :param weights: The list of weights of interaction to consider
    :param global_weights: The list global weights to consider. if \
    the global weight is equal to 0 then then density figure are calculated \
    by project, else all projcet are merge together and the interaction \
    seen in `global_weight` project are taken into account
    :param same_gene: Say if we consider as co-localised exon within the \
    same gene
    :param feature: The kind of analysed feature
    :param region: the region of interest to extract from gene
    :param logging_level: Level of information to display
    """
    ConfigGraph.community_folder.mkdir(exist_ok=True, parents=True)
    logging_def(ConfigGraph.community_folder, __file__, logging_level)
    logging.info("Checking if communities as an effect on nucleotide "
                 "frequency")
    global_weights = list(np.unique(global_weights))
    weights = list(np.unique(weights))
    projects, dic_project = get_projects_name(global_weights)
    nt_list = ["A", "C", "G", "T", "S", "W"]
    condition = list(product(projects, weights, nt_list))
    processes = {}
    pool = mp.Pool(processes=min(ps, len(condition)))
    logging.debug("Calculating stats...")
    dic_df = {}
    for project, weight, nt in condition:
        global_weight = dic_project[project]
        ckey = get_key(project, weight)
        if ckey in dic_df:
            df = dic_df[ckey]
        else:
            df = create_dataframe(project, weight, global_weight, same_gene,
                                  feature)
            nfile_table = ConfigGraph.get_community_file(project, weight,
                                                         global_weight,
                                                         same_gene, feature,
                                                         f"_nt_table.txt",
                                                         True)
            df.to_csv(nfile_table, sep="\t", index=False)
            dic_df[ckey] = df
        args = [df, project, weight, global_weight, same_gene, nt, feature,
                region]
        if ckey not in processes.keys():
            processes[ckey] = [pool.apply_async(get_stat_nt_communities, args)]
        else:
            processes[ckey].append(
                pool.apply_async(get_stat_nt_communities, args))
    for p, value in processes.items():
        project, weight = p.split("_")
        results = [p.get(timeout=None) for p in value]
        pool.close()
        pool.join()
        fdf = pd.DataFrame(results)
        fdf["padj"] = multipletests(fdf['pval'].values, method='fdr_bh')[1]
        outfile = ConfigGraph.get_community_file(project, weight,
                                                 dic_project[project],
                                                 same_gene, feature,
                                                 f"lmm-nt_stat.txt",
                                                 True)
        nfolder = outfile.parent / "nt_analysis"
        noutfile = nfolder / outfile.name
        fdf.to_csv(noutfile, sep="\t", index=False)


if __name__ == "__main__":
    testmod()
