#!/usr/bin/env python3

# -*- coding: UTF-8 -*-

"""
Description: The goal of this script is to get only coding gene from \
ensemble with their protein id and their coordinate in hg19.
"""

from .config_ppi import ConfigPPI
import pandas as pd
from pathlib import Path
import numpy as np


def load_genes(ensembl_file: Path) -> pd.DataFrame:
    """
    Load a table containing all the human protein of ensembl database \
    and the name and location of their genes.

    :param ensembl_file: A file containing all the human protein of \
    ensembl database and the name and location of their genes.
    :return: table containing all the human protein of ensembl database \
    and the name and location of their genes.
    """
    df = pd.read_csv(ensembl_file, sep="\t",
                     dtype={"Gene_stable_ID": str,
                            "Gene_stable_ID_version": str,
                            "Protein_stable_ID": str,
                            "Chromosome-scaffold_name": str,
                            "Gene_start": int,
                            "Gene_end": int,
                            "Strand": str,
                            "Gene_name": str,
                            "Gene_Synonym": str})
    df = df.loc[-df['Protein_stable_ID'].isna(), :]
    return df[["Gene_stable_ID", "Protein_stable_ID", "Gene_name",
               "Gene_Synonym"]]


def load_bed(ensembl_bed: Path) -> pd.DataFrame:
    """
    Load a bed file containing ensembl hg38 gene on hg19 coordinates.

    :param ensembl_bed: A bed file containing ensembl gene
    :return: the bed file
    """
    df = pd.read_csv(ensembl_bed, names=["chr", "start", "stop", "name",
                                         "score", "strand"], sep="\t")
    df['chr'] = df["chr"].str.replace("chr", "")
    df = df.drop_duplicates()
    names = np.unique(df['name'].values, return_counts=True)
    good_names = [cname for i, cname in enumerate(list(names[0]))
                  if names[1][i] == 1]
    return df.loc[df['name'].isin(good_names), :]


def merge_tables(df_gene: pd.DataFrame, df_bed: pd.DataFrame) -> pd.DataFrame:
    """
    Merges `df_gene` and `df_bed`together.

    :param df_gene: table containing all the human protein of ensembl \
     database and the name and location of their genes.
    :param df_bed: A bed file containing ensembl gene
    :return: The merged table
    """
    df_gene = df_gene.loc[
        df_gene["Gene_name"].isin(list(df_bed['name'].unique())), :]
    df_bed.rename({'name': "Gene_name"}, inplace=True, axis=1)
    return df_gene.merge(df_bed, how="left", on="Gene_name")


def create_protein_gene_file():
    """
    Create a file linking each ensembl protein to it's gene and the hg19 \
    coordinate of this gene.
    """

    df_gene = load_genes(ConfigPPI.ensembl_gene)
    df_bed = load_bed(ConfigPPI.bed_ensembl)
    df = merge_tables(df_gene, df_bed)
    ConfigPPI.protein_gene.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(ConfigPPI.protein_gene, sep="\t", index=False)
